"""Events by buildings routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

EVENTS_BY_BUILDINGS_URL = "/events_by_buildings/"
EVENTS_URL = "/events/"


class TestEventByBuildingApi:
    def test_events_by_buildings_api(self, app, users, buildings, events):

        creds = users["Chuck"]["creds"]
        event_1_id = events[0]
        event_2_id = events[1]
        building_1_id = buildings[0]
        building_2_id = buildings[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(EVENTS_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            ebb_1 = {
                "event_id": event_1_id,
                "building_id": building_1_id,
            }
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb_1)
            assert ret.status_code == 201
            ret_val = ret.json
            ebb_1_id = ret_val.pop("id")
            ebb_1_etag = ret.headers["ETag"]
            assert ret_val == ebb_1

            # POST violating unique constraint
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb_1)
            assert ret.status_code == 409

            # POST event + building from different campaigns
            ebb = {
                "event_id": event_2_id,
                "building_id": building_1_id,
            }
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb)
            assert ret.status_code == 422
            ret_val = ret.json
            assert ret_val["errors"]["json"]["_schema"] == (
                "Event and building must be in same campaign"
            )

            # GET list
            ret = client.get(EVENTS_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == building_1_id

            # GET by id
            ret = client.get(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == ebb_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == ebb_1

            # POST
            ebb_2 = {
                "event_id": event_2_id,
                "building_id": building_2_id,
            }
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb_2)
            ret_val = ret.json
            ebb_2_id = ret_val.pop("id")

            # GET list
            ret = client.get(EVENTS_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                EVENTS_BY_BUILDINGS_URL,
                query_string={"event_id": event_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ebb_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(f"{EVENTS_BY_BUILDINGS_URL}{DUMMY_ID}")
            assert ret.status_code == 404

            # DELETE event cascade
            ret = client.get(f"{EVENTS_URL}{event_1_id}")
            event_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{EVENTS_URL}{event_1_id}", headers={"If-Match": event_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
            assert ret.status_code == 404
            ret = client.delete(f"{EVENTS_BY_BUILDINGS_URL}{ebb_2_id}")
            assert ret.status_code == 204

            # GET list
            ret = client.get(EVENTS_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_events_by_buildings_as_user_api(
        self, app, users, buildings, events, events_by_buildings
    ):

        user_creds = users["Active"]["creds"]
        event_1_id = events[0]
        event_2_id = events[1]
        building_1_id = buildings[0]
        building_2_id = buildings[1]
        ebb_1_id = events_by_buildings[0]
        ebb_2_id = events_by_buildings[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(EVENTS_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ebb_1 = ret_val[0]
            assert ebb_1.pop("id") == ebb_1_id

            # GET by id
            ret = client.get(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
            assert ret.status_code == 200

            # GET by id not in campaign scope
            ret = client.get(f"{EVENTS_BY_BUILDINGS_URL}{ebb_2_id}")
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
            assert ret.status_code == 204

            # POST
            ebb_3 = {
                "event_id": event_1_id,
                "building_id": building_1_id,
            }
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb_3)
            assert ret.status_code == 201
            ret_val = ret.json

            # POST not in campaign scope
            ebb = {
                "event_id": event_2_id,
                "building_id": building_2_id,
            }
            ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb)
            assert ret.status_code == 403

    def test_events_by_buildings_as_anonym_api(
        self, app, buildings, events, events_by_buildings
    ):
        event_1_id = events[0]
        building_2_id = buildings[1]
        ebb_1_id = events_by_buildings[0]

        client = app.test_client()

        # GET list
        ret = client.get(EVENTS_BY_BUILDINGS_URL)
        assert ret.status_code == 401

        # POST
        ebb_3 = {
            "event_id": event_1_id,
            "building_id": building_2_id,
        }
        ret = client.post(EVENTS_BY_BUILDINGS_URL, json=ebb_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(f"{EVENTS_BY_BUILDINGS_URL}{ebb_1_id}")
        assert ret.status_code == 401



"""Completeness API schemas"""
import marshmallow as ma

from bemserver_api import Schema
from bemserver_api.extensions.ma_fields import Timezone
from bemserver_api.resources.timeseries_data.schemas import TimeseriesBucketWidthSchema


class TimeseriesCompletenessSchema(Schema):
    name = ma.fields.String(
        metadata={
            "description": "Timeseries name",
        },
    )
    count = ma.fields.List(
        ma.fields.Integer(),
        metadata={
            "description": "Number or values for each bucket",
        },
    )
    ratio = ma.fields.List(
        ma.fields.Float(allow_none=True),
        metadata={
            "description": "Number or values / expected values for each bucket",
        },
    )
    total_count = ma.fields.Integer(
        metadata={
            "description": "Total number or values",
        },
    )
    avg_count = ma.fields.Float(
        metadata={
            "description": "Average number or values",
        },
    )
    avg_ratio = ma.fields.Float(
        allow_none=True,
        metadata={
            "description": "Average number or values / expected values",
        },
    )
    interval = ma.fields.Float(
        allow_none=True,
        metadata={
            "description": "Interval between samples (seconds), defined or inferred",
        },
    )
    undefined_interval = ma.fields.Boolean(
        metadata={
            "description": "Whether the interval was not defined",
        },
    )
    expected_count = ma.fields.List(
        ma.fields.Float(allow_none=True),
        metadata={
            "description": "Expected number or values for each bucket",
        },
    )


class CompletenessSchema(Schema):
    timestamps = ma.fields.List(
        ma.fields.AwareDateTime,
        metadata={
            "description": "Time index (value is bucket start time)",
        },
    )
    timeseries = ma.fields.Dict(
        keys=ma.fields.String(metadata={"description": "LOL"}),
        values=ma.fields.Nested(TimeseriesCompletenessSchema),
    )


class CompletenessQueryArgsSchema(TimeseriesBucketWidthSchema):
    start_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "Initial datetime",
        },
    )
    end_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "End datetime (excluded from the interval)",
        },
    )
    timeseries = ma.fields.List(
        ma.fields.Int(),
        required=True,
        metadata={
            "description": "List of timeseries ID",
        },
    )
    data_state = ma.fields.Int(
        required=True,
        metadata={
            "description": "Data state ID",
        },
    )
    timezone = Timezone(
        load_default="UTC",
        metadata={
            "description": "Timezone to use for the aggreagation",
        },
    )



"""Events by zones resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import EventByZone
from bemserver_core.exceptions import BEMServerCoreCampaignError

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    EventByZoneSchema,
    EventByZoneQueryArgsSchema,
)


blp = Blueprint(
    "EventByZone",
    __name__,
    url_prefix="/events_by_zones",
    description="Operations on event x zone associations",
)


@blp.route("/")
class EventByZoneViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(EventByZoneQueryArgsSchema, location="query")
    @blp.response(200, EventByZoneSchema(many=True))
    def get(self, args):
        """List event x zone associations"""
        return EventByZone.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(EventByZoneSchema)
    @blp.response(201, EventByZoneSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new event x zone association"""
        item = EventByZone.new(**new_item)
        try:
            db.session.commit()
        except BEMServerCoreCampaignError as exc:
            abort(422, errors={"json": {"_schema": str(exc)}})
        return item


@blp.route("/<int:item_id>")
class EventByZoneByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EventByZoneSchema)
    def get(self, item_id):
        """Get event x zone association ID"""
        item = EventByZone.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.response(204)
    def delete(self, item_id):
        """Delete an event x zone association"""
        item = EventByZone.get_by_id(item_id)
        if item is None:
            abort(404)
        item.delete()
        db.session.commit()



"""Timeseries resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Timeseries

from bemserver_api import Blueprint, SQLCursorPage
from bemserver_api.database import db

from .schemas import TimeseriesSchema, TimeseriesPutSchema, TimeseriesQueryArgsSchema


blp = Blueprint(
    "Timeseries",
    __name__,
    url_prefix="/timeseries",
    description="Operations on timeseries",
)


@blp.route("/")
class TimeseriesViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesQueryArgsSchema, location="query")
    @blp.response(200, TimeseriesSchema(many=True))
    @blp.paginate(SQLCursorPage)
    def get(self, args):
        """List timeseries"""
        return Timeseries.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesSchema)
    @blp.response(201, TimeseriesSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new timeseries"""
        item = Timeseries.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class TimeseriesByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, TimeseriesSchema)
    def get(self, item_id):
        """Get timeseries by ID"""
        item = Timeseries.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesPutSchema)
    @blp.response(200, TimeseriesSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing timeseries"""
        item = Timeseries.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a timeseries"""
        item = Timeseries.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesSchema)
        item.delete()
        db.session.commit()



"""ST_CleanupByTimeseries API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.scheduled_tasks import ST_CleanupByTimeseries

from bemserver_api import AutoSchema, Schema, SortField


class ST_CleanupByTimeseriesSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = ST_CleanupByTimeseries

    id = msa.auto_field(dump_only=True)


class ST_CleanupByTimeseriesFullSchema(Schema):
    id = ma.fields.Int()
    timeseries_id = ma.fields.Int()
    timeseries_name = ma.fields.Str()
    timeseries_unit_symbol = ma.fields.Str()
    last_timestamp = ma.fields.AwareDateTime(allow_none=True)


class ST_CleanupByTimeseriesQueryArgsSchema(Schema):
    campaign_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()


class ST_CleanupByTimeseriesFullQueryArgsSchema(Schema):
    sort = SortField(
        (
            "timeseries_name",
            "last_timestamp",
        )
    )
    campaign_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()
    in_timeseries_name = ma.fields.Str(
        metadata={
            "description": (
                "Search for items whose name contains this input value"
                " (case insensitive)"
            ),
        }
    )



"""Completeness tests"""
import contextlib

import pytest

from tests.common import AuthHeader


COMPLETENESS_URL = "/analysis/completeness"


class TestAnalysisApiCompleteness:
    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_analysis_completeness(
        self,
        app,
        user,
        users,
        timeseries,
        timeseries_data,
    ):
        start_time, end_time = timeseries_data
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        ds_id = 1

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        with auth_context:

            query_url = COMPLETENESS_URL
            ts_l = (ts_1_id,)

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == {
                    "timestamps": [
                        "2020-01-01T00:00:00+00:00",
                        "2020-01-01T01:00:00+00:00",
                        "2020-01-01T02:00:00+00:00",
                        "2020-01-01T03:00:00+00:00",
                    ],
                    "timeseries": {
                        "1": {
                            "name": "Timeseries 0",
                            "avg_count": 1.0,
                            "avg_ratio": 1.0,
                            "count": [1, 1, 1, 1],
                            "expected_count": [1.0, 1.0, 1.0, 1.0],
                            "interval": 3600.0,
                            "ratio": [1.0, 1.0, 1.0, 1.0],
                            "total_count": 4,
                            "undefined_interval": True,
                        }
                    },
                }

            # User not in Timeseries group
            query_url = COMPLETENESS_URL
            ts_l = (ts_2_id,)

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == {
                    "timestamps": [
                        "2020-01-01T00:00:00+00:00",
                        "2020-01-01T01:00:00+00:00",
                        "2020-01-01T02:00:00+00:00",
                        "2020-01-01T03:00:00+00:00",
                    ],
                    "timeseries": {
                        "2": {
                            "name": "Timeseries 1",
                            "avg_count": 1.0,
                            "avg_ratio": 1.0,
                            "count": [1, 1, 1, 1],
                            "expected_count": [1.0, 1.0, 1.0, 1.0],
                            "interval": 3600.0,
                            "ratio": [1.0, 1.0, 1.0, 1.0],
                            "total_count": 4,
                            "undefined_interval": True,
                        }
                    },
                }



"""Events by sites API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import EventBySite

from bemserver_api import AutoSchema, Schema


class EventBySiteSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = EventBySite

    id = msa.auto_field(dump_only=True)


class EventBySiteQueryArgsSchema(Schema):
    site_id = ma.fields.Int()
    event_id = ma.fields.Int()



"""Input/Output resources"""
import io

from flask_smorest import abort

from bemserver_core.model import Campaign
from bemserver_core.input_output import sites_csv_io, timeseries_csv_io
from bemserver_core.exceptions import BEMServerCoreIOError

from bemserver_api import Blueprint

from .schemas import (
    SitesCSVUploadQueryArgsSchema,
    SitesCSVUploadFileSchema,
    TimeseriesCSVUploadQueryArgsSchema,
    TimeseriesCSVUploadFileSchema,
)


blp = Blueprint(
    "Input/Output", __name__, url_prefix="/io", description="Input/Output operations"
)


@blp.post("/sites")
@blp.login_required
@blp.arguments(SitesCSVUploadQueryArgsSchema, location="query")
@blp.arguments(SitesCSVUploadFileSchema, location="files")
@blp.response(201)
def sites_csv_io_post(args, files):
    """Import site description tree"""
    campaign_id = args["campaign_id"]
    csv_files = {
        k: io.TextIOWrapper(v, encoding="utf-8")
        for k, v in files.items()
        if v is not None
    }

    campaign = Campaign.get_by_id(campaign_id)
    if campaign is None:
        abort(422, errors={"query": {"campaign_id": ["Unknown campaign ID."]}})
    try:
        sites_csv_io.import_csv(campaign=campaign, **csv_files)
    except BEMServerCoreIOError as exc:
        abort(422, message=f"Invalid CSV file content: {exc}")


@blp.post("/timeseries")
@blp.login_required
@blp.arguments(TimeseriesCSVUploadQueryArgsSchema, location="query")
@blp.arguments(TimeseriesCSVUploadFileSchema, location="files")
@blp.response(201)
def timeseries_csv_io_post(args, files):
    """Import timeseries description tree"""
    campaign_id = args["campaign_id"]
    timeseries_csv = io.TextIOWrapper(files["timeseries_csv"], encoding="utf-8")

    campaign = Campaign.get_by_id(campaign_id)
    if campaign is None:
        abort(422, errors={"query": {"campaign_id": ["Unknown campaign ID."]}})
    try:
        timeseries_csv_io.import_csv(campaign=campaign, timeseries_csv=timeseries_csv)
    except BEMServerCoreIOError as exc:
        abort(422, message=f"Invalid CSV file content: {exc}")



"""Users by user groups API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import UserByUserGroup

from bemserver_api import AutoSchema, Schema


class UserByUserGroupSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = UserByUserGroup

    id = msa.auto_field(dump_only=True)


class UserByUserGroupQueryArgsSchema(Schema):
    user_id = ma.fields.Int()
    user_group_id = ma.fields.Int()



"""Integrity error management

Catch integrity errors in resources and return appropriate error code.
"""
import contextlib

import sqlalchemy as sqla
import psycopg2.errors as ppe
from flask_smorest import abort


class catch_integrity_error(contextlib.ContextDecorator):
    """Context manager catching integrity errors

    Can be used as context manager or decorator.
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type and issubclass(exc_type, sqla.exc.IntegrityError):
            if isinstance(exc_value.orig, ppe.UniqueViolation):
                abort(409, "Unique constraint violation")
            if isinstance(exc_value.orig, ppe.ForeignKeyViolation):
                abort(409, "Foreign key constraint violation")
            # Shouldn't happen
            abort(409)
        return False



"""Energy consumption tests"""
import datetime as dt
import contextlib

import pandas as pd

import pytest

from bemserver_core.database import db
from bemserver_core.model import (
    UserGroupByCampaignScope,
    Timeseries,
    TimeseriesDataState,
    EnergySource,
    EnergyEndUse,
    EnergyConsumptionTimeseriesBySite,
    EnergyConsumptionTimeseriesByBuilding,
)
from bemserver_core.authorization import OpenBar

from tests.common import AuthHeader
from tests.utils import create_timeseries_data


ENERGY_CONSUMPTION_URL = "/analysis/energy_consumption/"


class TestAnalysisApiEnergyConsumption:
    def _create_data(self, campaign_id, campaign_scope_id):
        start_dt = dt.datetime(2020, 1, 1, 0, 0, tzinfo=dt.timezone.utc)
        end_dt = dt.datetime(2020, 1, 1, 2, 0, tzinfo=dt.timezone.utc)

        timestamps = pd.date_range(start_dt, end_dt, inclusive="left", freq="H")

        # Create timeseries and timeseries data
        ds_clean = TimeseriesDataState.get(name="Clean").first()

        timeseries = []
        for i in range(9):
            ts_i = Timeseries(
                name=f"Timeseries {i+1}",
                campaign_id=campaign_id,
                campaign_scope_id=campaign_scope_id,
            )
            timeseries.append(ts_i)
        db.session.add_all(timeseries)
        db.session.flush()

        create_timeseries_data(timeseries[0], ds_clean, timestamps, [71, 71])
        create_timeseries_data(timeseries[1], ds_clean, timestamps, [46, 46])
        create_timeseries_data(timeseries[2], ds_clean, timestamps, [25, 25])
        create_timeseries_data(timeseries[3], ds_clean, timestamps, [50, 50])
        create_timeseries_data(timeseries[4], ds_clean, timestamps, [25, 25])
        create_timeseries_data(timeseries[5], ds_clean, timestamps, [25, 25])
        create_timeseries_data(timeseries[6], ds_clean, timestamps, [0.021, 0.021])
        create_timeseries_data(timeseries[7], ds_clean, timestamps, [0.021, 0.021])

        expected_consumptions = {
            "all": {
                "all": [71.0, 71.0],
                "heating": [46.0, 46.0],
                "cooling": [25.0, 25.0],
            },
            "electricity": {
                "all": [50.0, 50.0],
                "heating": [25.0, 25.0],
                "cooling": [25.0, 25.0],
            },
            "natural gas": {
                "all": [21.0, 21.0],
                "heating": [21.0, 21.0],
                "cooling": [0.0, 0.0],
            },
        }

        expected = {
            "timestamps": [ts.isoformat() for ts in timestamps.to_list()],
            "energy": expected_consumptions,
        }

        return start_dt, end_dt, timeseries, expected

    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_analysis_energy_consumption_breakdown_for_site(
        self,
        app,
        user,
        users,
        campaigns,
        campaign_scopes,
        sites,
        timeseries,
    ):
        campaign_1_id = campaigns[0]
        cs_1_id = campaign_scopes[0]
        site_1_id = sites[0]

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        query_url = f"{ENERGY_CONSUMPTION_URL}site/{site_1_id}"

        with OpenBar():

            start_time, end_time, timeseries, expected = self._create_data(
                campaign_1_id, cs_1_id
            )

            source_all = EnergySource.get(name="all").first()
            source_elec = EnergySource.get(name="electricity").first()
            source_gas = EnergySource.get(name="natural gas").first()
            end_use_all = EnergyEndUse.get(name="all").first()
            end_use_heating = EnergyEndUse.get(name="heating").first()
            end_use_cooling = EnergyEndUse.get(name="cooling").first()

            ectbs_all_all = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_all.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[0].id,
            )
            ectbs_all_heating = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_all.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[1].id,
            )
            ectbs_all_cooling = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_all.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[2].id,
            )
            ectbs_elec_all = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_elec.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[3].id,
            )
            ectbs_elec_heating = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_elec.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[4].id,
            )
            ectbs_elec_cooling = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_elec.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[5].id,
            )
            ectbs_gas_all = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_gas.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[6].id,
                wh_conversion_factor=1000,
            )
            ectbs_gas_heating = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_gas.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[7].id,
                wh_conversion_factor=1000,
            )
            ectbs_gas_cooling = EnergyConsumptionTimeseriesBySite.new(
                site_id=sites[0],
                source_id=source_gas.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[8].id,
                wh_conversion_factor=1000,
            )
            db.session.add_all(
                (
                    ectbs_all_all,
                    ectbs_all_heating,
                    ectbs_all_cooling,
                    ectbs_elec_all,
                    ectbs_elec_heating,
                    ectbs_elec_cooling,
                    ectbs_gas_all,
                    ectbs_gas_heating,
                    ectbs_gas_cooling,
                )
            )
            db.session.commit()

        with auth_context:

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected

        with OpenBar():
            # Remove user <-> timeseries association
            ugbcs_1 = UserGroupByCampaignScope.get(campaign_scope_id=cs_1_id).first()
            ugbcs_1.delete()
            db.session.commit()

        with auth_context:

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                # User can't read consumption data from unauthorized timeseries
                expected["energy"] = {}
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected

    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_analysis_energy_consumption_breakdown_for_building(
        self,
        app,
        user,
        users,
        campaigns,
        campaign_scopes,
        buildings,
        timeseries,
    ):
        campaign_1_id = campaigns[0]
        cs_1_id = campaign_scopes[0]
        building_1_id = buildings[0]

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        query_url = f"{ENERGY_CONSUMPTION_URL}building/{building_1_id}"

        with OpenBar():

            start_time, end_time, timeseries, expected = self._create_data(
                campaign_1_id, cs_1_id
            )

            source_all = EnergySource.get(name="all").first()
            source_elec = EnergySource.get(name="electricity").first()
            source_gas = EnergySource.get(name="natural gas").first()
            end_use_all = EnergyEndUse.get(name="all").first()
            end_use_heating = EnergyEndUse.get(name="heating").first()
            end_use_cooling = EnergyEndUse.get(name="cooling").first()

            ectbs_all_all = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_all.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[0].id,
            )
            ectbs_all_heating = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_all.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[1].id,
            )
            ectbs_all_cooling = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_all.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[2].id,
            )
            ectbs_elec_all = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_elec.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[3].id,
            )
            ectbs_elec_heating = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_elec.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[4].id,
            )
            ectbs_elec_cooling = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_elec.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[5].id,
            )
            ectbs_gas_all = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_gas.id,
                end_use_id=end_use_all.id,
                timeseries_id=timeseries[6].id,
                wh_conversion_factor=1000,
            )
            ectbs_gas_heating = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_gas.id,
                end_use_id=end_use_heating.id,
                timeseries_id=timeseries[7].id,
                wh_conversion_factor=1000,
            )
            ectbs_gas_cooling = EnergyConsumptionTimeseriesByBuilding.new(
                building_id=buildings[0],
                source_id=source_gas.id,
                end_use_id=end_use_cooling.id,
                timeseries_id=timeseries[8].id,
                wh_conversion_factor=1000,
            )
            db.session.add_all(
                (
                    ectbs_all_all,
                    ectbs_all_heating,
                    ectbs_all_cooling,
                    ectbs_elec_all,
                    ectbs_elec_heating,
                    ectbs_elec_cooling,
                    ectbs_gas_all,
                    ectbs_gas_heating,
                    ectbs_gas_cooling,
                )
            )
            db.session.commit()

        with auth_context:

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected

        with OpenBar():
            # Remove user <-> timeseries association
            ugbcs_1 = UserGroupByCampaignScope.get(campaign_scope_id=cs_1_id).first()
            ugbcs_1.delete()
            db.session.commit()

        with auth_context:

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "bucket_width_value": 1,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                # User can't read consumption data from unauthorized timeseries
                expected["energy"] = {}
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected
            else:
                assert ret.status_code == 200
                ret_data = ret.json
                assert ret_data == expected



"""REST API extension"""
from copy import deepcopy
from functools import wraps
import http

import marshmallow as ma
import flask_smorest
from webargs.fields import DelimitedList
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.common import resolve_schema_cls
import marshmallow_sqlalchemy as msa

from .ma_fields import Timezone
from .authentication import auth
from . import integrity_error


def resolver(schema):
    # This is the default name resolver from apispec
    schema_cls = resolve_schema_cls(schema)
    name = schema_cls.__name__
    if name.endswith("Schema"):
        name = name[:-6] or name
    # Except it appends ExcludeId to schemas where id is excluded.
    if isinstance(schema, ma.Schema) and "id" in schema.exclude:
        name = f"{name}ExcludeId"
    return name


class Api(flask_smorest.Api):
    """Api class"""

    def __init__(self, app=None, *, spec_kwargs=None):
        spec_kwargs = spec_kwargs or {}
        spec_kwargs["marshmallow_plugin"] = MarshmallowPlugin(
            schema_name_resolver=resolver
        )
        super().__init__(app=app, spec_kwargs=spec_kwargs)

    def init_app(self, app, *, spec_kwargs=None):
        super().init_app(app, spec_kwargs=spec_kwargs)
        self.register_field(Timezone, "string", "iana-tz")
        self.spec.components.security_scheme(
            "BasicAuthentication", {"type": "http", "scheme": "basic"}
        )


class Blueprint(flask_smorest.Blueprint):
    """Blueprint class"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prepare_doc_cbks.append(self._prepare_auth_doc)

    @staticmethod
    def login_required(func=None, **kwargs):
        def decorator(function):
            function = auth.login_required(**kwargs)(function)
            getattr(function, "_apidoc", {})["auth"] = True
            return function

        if func is None:
            return decorator
        return decorator(func)

    @staticmethod
    def _prepare_auth_doc(doc, doc_info, *, app, **kwargs):
        if doc_info.get("auth", False):
            doc.setdefault("responses", {})["401"] = http.HTTPStatus(401).name
            doc["security"] = [{"BasicAuthentication": []}]
        return doc

    @staticmethod
    def current_user():
        return auth.current_user()

    @staticmethod
    def catch_integrity_error(func=None):
        """Catch DB integrity errors"""

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                with integrity_error.catch_integrity_error():
                    return func(*args, **kwargs)

            # Store doc in wrapper function
            # The deepcopy avoids modifying the wrapped function doc
            wrapper._apidoc = deepcopy(getattr(wrapper, "_apidoc", {}))
            wrapper._apidoc.setdefault("response", {}).setdefault("responses", {})[
                409
            ] = http.HTTPStatus(409).name

            return wrapper

        if func is None:
            return decorator
        return decorator(func)


class Schema(ma.Schema):
    """Base Schema class to use in the API"""

    # Ensures the fields are ordered
    set_class = ma.orderedset.OrderedSet


class AutoSchema(msa.SQLAlchemyAutoSchema, Schema):
    """Auto Schema class used to load/dump SQLAlchemy objects

    The API assumes missing = None/null, so we treat missing fields as None
    unless they are read-only.
    """

    class Meta:
        include_fk = True

    @ma.post_load
    def set_missing_expected_values_to_none(self, data, **kwargs):
        loadable_fields = [k for k, v in self.fields.items() if not v.dump_only]
        for name in loadable_fields:
            data.setdefault(name, None)
        return data

    @ma.post_dump
    def remove_none_values(self, data, **kwargs):
        return {key: value for key, value in data.items() if value is not None}


class SQLCursorPage(flask_smorest.Page):
    """SQL cursor pager"""

    @property
    def item_count(self):
        return self.collection.count()


class SortField(DelimitedList):
    """Field used to specify sort order fields

    :param list fields: List of fields to sort upon, by order of priority (the
    first field is the first sort key). Each field is a field name, optionally
    prefixed with "+" or "-".
    """

    def __init__(self, fields, **kwargs):
        validator = ma.validate.OneOf(
            [v for f in fields for v in [f, f"+{f}", f"-{f}"]]
        )
        super().__init__(ma.fields.String(validate=validator), **kwargs)



"""User groups by campaign scopes API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import UserGroupByCampaignScope

from bemserver_api import AutoSchema, Schema


class UserGroupByCampaignScopeSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = UserGroupByCampaignScope

    id = msa.auto_field(dump_only=True)


class UserGroupByCampaignScopeQueryArgsSchema(Schema):
    user_group_id = ma.fields.Int()
    campaign_scope_id = ma.fields.Int()



"""Storeys routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

STOREYS_URL = "/storeys/"


class TestStoreysApi:
    def test_storeys_api(self, app, users, campaigns, sites, buildings):

        creds = users["Chuck"]["creds"]
        campaign_2_id = campaigns[1]
        site_2_id = sites[1]
        building_1_id = buildings[0]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(STOREYS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            storey_1 = {
                "name": "Storey 1",
                "building_id": building_1_id,
            }
            ret = client.post(STOREYS_URL, json=storey_1)
            assert ret.status_code == 201
            ret_val = ret.json
            storey_1_id = ret_val.pop("id")
            storey_1_etag = ret.headers["ETag"]
            assert ret_val == storey_1

            # POST violating unique constraint
            ret = client.post(STOREYS_URL, json=storey_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(STOREYS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == storey_1_id

            # GET by id
            ret = client.get(f"{STOREYS_URL}{storey_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == storey_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == storey_1

            # PUT
            storey_1["description"] = "Fantastic storey"
            storey_1_put = storey_1.copy()
            del storey_1_put["building_id"]
            ret = client.put(
                f"{STOREYS_URL}{storey_1_id}",
                json=storey_1_put,
                headers={"If-Match": storey_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            storey_1_etag = ret.headers["ETag"]
            assert ret_val == storey_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{STOREYS_URL}{DUMMY_ID}",
                json=storey_1_put,
                headers={"If-Match": storey_1_etag},
            )
            assert ret.status_code == 404

            # POST storey 2
            storey_2 = {
                "name": "Storey 2",
                "building_id": building_1_id,
            }
            ret = client.post(STOREYS_URL, json=storey_2)
            ret_val = ret.json
            storey_2_id = ret_val.pop("id")
            storey_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            storey_1_put["name"] = storey_2["name"]
            ret = client.put(
                f"{STOREYS_URL}{storey_1_id}",
                json=storey_1_put,
                headers={"If-Match": storey_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(STOREYS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(STOREYS_URL, query_string={"name": "Storey 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == storey_1_id
            ret = client.get(STOREYS_URL, query_string={"campaign_id": campaign_2_id})
            assert ret.status_code == 200
            ret_val = ret.json
            assert not ret_val
            ret = client.get(STOREYS_URL, query_string={"site_id": site_2_id})
            assert ret.status_code == 200
            ret_val = ret.json
            assert not ret_val

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{STOREYS_URL}{DUMMY_ID}", headers={"If-Match": storey_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{STOREYS_URL}{storey_1_id}", headers={"If-Match": storey_1_etag}
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{STOREYS_URL}{storey_2_id}", headers={"If-Match": storey_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(STOREYS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{STOREYS_URL}{storey_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_storeys_as_user_api(self, app, users, buildings, storeys):

        user_creds = users["Active"]["creds"]
        building_2_id = buildings[1]
        storey_1_id = storeys[0]
        storey_2_id = storeys[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(STOREYS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            storey_1 = ret_val[0]
            assert storey_1.pop("id") == storey_1_id

            # GET list with filters
            ret = client.get(STOREYS_URL, query_string={"name": "Storey 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == storey_1_id
            ret = client.get(STOREYS_URL, query_string={"name": "Storey 2"})
            assert ret.status_code == 200
            assert not ret.json

            # POST
            storey_3 = {
                "name": "Storey 3",
                "building_id": building_2_id,
            }
            ret = client.post(STOREYS_URL, json=storey_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{STOREYS_URL}{storey_1_id}")
            assert ret.status_code == 200
            storey_1_etag = ret.headers["ETag"]

            ret = client.get(f"{STOREYS_URL}{storey_2_id}")
            assert ret.status_code == 403

            # PUT
            storey_1["description"] = "Fantastic storey"
            storey_1_put = storey_1.copy()
            del storey_1_put["building_id"]
            ret = client.put(
                f"{STOREYS_URL}{storey_1_id}",
                json=storey_1_put,
                headers={"If-Match": storey_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{STOREYS_URL}{storey_1_id}", headers={"If-Match": storey_1_etag}
            )
            assert ret.status_code == 403

    def test_storeys_as_anonym_api(self, app, storeys, buildings):

        storey_1_id = storeys[0]
        building_1_id = buildings[0]

        client = app.test_client()

        # GET list
        ret = client.get(STOREYS_URL)
        assert ret.status_code == 401

        # POST
        storey_3 = {
            "name": "Storey 3",
            "building_id": building_1_id,
        }
        ret = client.post(STOREYS_URL, json=storey_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{STOREYS_URL}{storey_1_id}")
        assert ret.status_code == 401

        # PUT
        storey_1 = {
            "name": "Super Storey 1",
        }
        ret = client.put(
            f"{STOREYS_URL}{storey_1_id}",
            json=storey_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{STOREYS_URL}{storey_1_id}", headers={"If-Match": "Dummy-Etag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



from contextvars import ContextVar
from contextlib import AbstractContextManager

from bemserver_api.settings import Config


class TestConfig(Config):
    TESTING = True


AUTH_HEADER = ContextVar("auth_header", default=None)


class AuthHeader(AbstractContextManager):
    def __init__(self, creds):
        self.creds = creds

    def __enter__(self):
        self.token = AUTH_HEADER.set("Basic " + self.creds)

    def __exit__(self, *args, **kwargs):
        AUTH_HEADER.reset(self.token)



"""User groups by campaign scopes routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

USER_GROUPS_URL = "/user_groups/"
CAMPAIGN_SCOPES_URL = "/campaign_scopes/"
USER_GROUPS_BY_CAMPAIGN_SCOPES_URL = "/user_groups_by_campaign_scopes/"


class TestUsersByCampaignsApi:
    def test_user_groups_by_campaign_scopes_api(
        self, app, users, user_groups, campaign_scopes
    ):

        ug_1_id = user_groups[0]
        ug_2_id = user_groups[1]
        cs_1_id = campaign_scopes[0]
        cs_2_id = campaign_scopes[1]

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            ugbc_1 = {"campaign_scope_id": cs_1_id, "user_group_id": ug_1_id}
            ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_1)
            assert ret.status_code == 201
            ret_val = ret.json
            ugbc_1_id = ret_val.pop("id")
            ugbc_1_etag = ret.headers["ETag"]

            # POST violating unique constraint
            ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ugbc_1_id

            # GET by id
            ret = client.get(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == ugbc_1_etag

            # POST
            ugbc_2 = {"campaign_scope_id": cs_2_id, "user_group_id": ug_2_id}
            ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_2)
            assert ret.status_code == 201
            ret_val = ret.json
            ugbc_2_id = ret_val.pop("id")

            # GET list (filtered)
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"user_group_id": ug_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ugbc_1_id
            assert ret_val[0]["user_group_id"] == ug_1_id
            assert ret_val[0]["campaign_scope_id"] == cs_1_id
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"campaign_scope_id": cs_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ugbc_2_id
            assert ret_val[0]["user_group_id"] == ug_2_id
            assert ret_val[0]["campaign_scope_id"] == cs_2_id
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"user_group_id": ug_1_id, "campaign_scope_id": cs_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE wrong ID -> 404
            ret = client.delete(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{DUMMY_ID}")
            assert ret.status_code == 404

            # DELETE
            ugbc_3 = {"campaign_scope_id": cs_1_id, "user_group_id": ug_2_id}
            ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_3)
            assert ret.status_code == 201
            ret_val = ret.json
            ugbc_3_id = ret_val.pop("id")
            ret = client.delete(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_3_id}")
            assert ret.status_code == 204

            # DELETE user group cascade
            ret = client.get(f"{USER_GROUPS_URL}{ug_1_id}")
            ug_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{USER_GROUPS_URL}{ug_1_id}", headers={"If-Match": ug_1_etag}
            )
            assert ret.status_code == 204

            # DELETE campaign scope cascade
            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_2_id}")
            cs_2_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{CAMPAIGN_SCOPES_URL}{cs_2_id}", headers={"If-Match": cs_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    def test_user_groups_by_campaign_scopes_as_user_api(
        self, app, users, user_groups, campaign_scopes, user_groups_by_campaign_scopes
    ):

        user_creds = users["Active"]["creds"]
        ug_1_id = user_groups[0]
        ug_2_id = user_groups[1]
        cs_1_id = campaign_scopes[0]
        cs_2_id = campaign_scopes[1]
        ugbc_1_id = user_groups_by_campaign_scopes[0]
        ugbc_2_id = user_groups_by_campaign_scopes[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ugbc_1 = ret_val[0]
            assert ugbc_1["id"] == ugbc_1_id

            # POST
            ugbc_3 = {"campaign_scope_id": cs_1_id, "user_group_id": ug_1_id}
            ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
            assert ret.status_code == 200
            ret = client.get(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_2_id}")
            assert ret.status_code == 403

            # GET list (filtered)
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"user_group_id": ug_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ugbc_1_id
            assert ret_val[0]["user_group_id"] == ug_1_id
            assert ret_val[0]["campaign_scope_id"] == cs_1_id
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"campaign_scope_id": cs_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(
                USER_GROUPS_BY_CAMPAIGN_SCOPES_URL,
                query_string={"user_group_id": ug_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE
            ret = client.delete(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
            assert ret.status_code == 403

    def test_user_groups_by_campaign_scopes_as_anonym_api(
        self, app, users, user_groups, campaign_scopes, user_groups_by_campaign_scopes
    ):

        ug_1_id = user_groups[0]
        cs_1_id = campaign_scopes[0]
        ugbc_1_id = user_groups_by_campaign_scopes[0]

        client = app.test_client()

        # GET list
        ret = client.get(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL)
        assert ret.status_code == 401

        # POST
        ugbc_1 = {"campaign_scope_id": cs_1_id, "user_group_id": ug_1_id}
        ret = client.post(USER_GROUPS_BY_CAMPAIGN_SCOPES_URL, json=ugbc_1)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(f"{USER_GROUPS_BY_CAMPAIGN_SCOPES_URL}{ugbc_1_id}")
        assert ret.status_code == 401



"""Energy end uses API schemas"""

import marshmallow_sqlalchemy as msa

from bemserver_core.model import EnergyEndUse

from bemserver_api import AutoSchema


class EnergyEndUseSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = EnergyEndUse

    id = msa.auto_field(dump_only=True)



"""Campaigns API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Campaign

from bemserver_api import AutoSchema, Schema, SortField
from bemserver_api.extensions.ma_fields import Timezone


class CampaignSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Campaign

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    start_time = ma.fields.AwareDateTime()
    end_time = ma.fields.AwareDateTime()
    timezone = Timezone(
        required=True,
        metadata={
            "default": "UTC",
            "description": "Timezone used to localize timeseries in this campaign",
        },
    )


class CampaignQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    in_name = ma.fields.Str(
        metadata={
            "description": (
                "Search for items whose name contains this input value"
                " (case insensitive)"
            ),
        }
    )



"""Timeseries data states tests"""
from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_DATA_STATES_URL = "/timeseries_data_states/"


class TestTimeseriesDataStatesApi:
    def test_timeseries_data_states_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_DATA_STATES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert {x["name"] for x in ret_val} == {"Raw", "Clean"}

            # POST
            tds_1 = {
                "name": "Quality",
            }
            ret = client.post(TIMESERIES_DATA_STATES_URL, json=tds_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tds_1_id = ret_val.pop("id")
            tds_1_etag = ret.headers["ETag"]
            assert ret_val == tds_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_DATA_STATES_URL, json=tds_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_DATA_STATES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 3
            assert {x["name"] for x in ret_val} == {"Raw", "Clean", "Quality"}

            # GET by id
            ret = client.get(f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tds_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tds_1

            # PUT
            tds_1["name"] = "Qualität"
            ret = client.put(
                f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
                json=tds_1,
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            tds_1_etag = ret.headers["ETag"]
            assert ret_val == tds_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_DATA_STATES_URL}{DUMMY_ID}",
                json=tds_1,
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 404

            # POST TSDS 2
            tds_2 = {
                "name": "Accuracy",
            }
            ret = client.post(TIMESERIES_DATA_STATES_URL, json=tds_2)
            ret_val = ret.json
            tds_2_id = ret_val.pop("id")
            tds_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            tds_2["name"] = "Qualität"
            ret = client.put(
                f"{TIMESERIES_DATA_STATES_URL}{tds_2_id}",
                json=tds_2,
                headers={"If-Match": tds_2_etag},
            )
            assert ret.status_code == 409

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{TIMESERIES_DATA_STATES_URL}{tds_2_id}",
                headers={"If-Match": tds_2_etag},
            )

            # GET list
            ret = client.get(TIMESERIES_DATA_STATES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert {x["name"] for x in ret_val} == {"Raw", "Clean"}

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}")
            assert ret.status_code == 404

    def test_timeseries_data_states_as_user_api(self, app, users, timeseries):

        tds_1_id = 1

        creds = users["Active"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_DATA_STATES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert {x["name"] for x in ret_val} == {"Raw", "Clean"}

            # POST
            tds = {
                "name": "Quality",
            }
            ret = client.post(TIMESERIES_DATA_STATES_URL, json=tds)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}")
            assert ret.status_code == 200
            tds_1_etag = ret.headers["ETag"]
            ret_val = ret.json
            ret_val.pop("id")
            tds_1 = ret_val

            # PUT
            tds_1["name"] = "Qualität"
            ret = client.put(
                f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
                json=tds_1,
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 403

    def test_timeseries_data_states_as_anonym_api(self, app, users):

        tds_1_id = 1

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_DATA_STATES_URL)
        assert ret.status_code == 401

        # POST
        tds = {
            "name": "Quality",
        }
        ret = client.post(TIMESERIES_DATA_STATES_URL, json=tds)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}")
        assert ret.status_code == 401

        # PUT
        ret = client.put(
            f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
            json=tds,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_DATA_STATES_URL}{tds_1_id}",
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Energy end uses resources"""

from flask.views import MethodView

from bemserver_core.model import EnergyEndUse

from bemserver_api import Blueprint

from .schemas import EnergyEndUseSchema


blp = Blueprint(
    "EnergyEndUse",
    __name__,
    url_prefix="/energy_end_uses",
    description="Operations on energy end uses",
)


@blp.route("/")
class EnergyEndUseViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EnergyEndUseSchema(many=True))
    def get(self):
        """List energy end uses"""
        return EnergyEndUse.get()



"""Campaign scopes routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

CAMPAIGN_SCOPES_URL = "/campaign_scopes/"
CAMPAIGNS_URL = "/campaigns/"


class TestCampaignScopesApi:
    def test_campaign_scopes_api(self, app, users, campaigns):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            cs_1 = {
                "name": "Campaign scope 1",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(CAMPAIGN_SCOPES_URL, json=cs_1)
            assert ret.status_code == 201
            ret_val = ret.json
            cs_1_id = ret_val.pop("id")
            cs_1_etag = ret.headers["ETag"]
            assert ret_val == cs_1

            # POST violating unique constraint
            ret = client.post(CAMPAIGN_SCOPES_URL, json=cs_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == cs_1_id

            # GET by id
            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == cs_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == cs_1

            # PUT
            cs_1["description"] = "Fantastic campaign"
            cs_1_put = cs_1.copy()
            del cs_1_put["campaign_id"]
            ret = client.put(
                f"{CAMPAIGN_SCOPES_URL}{cs_1_id}",
                json=cs_1_put,
                headers={"If-Match": cs_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            cs_1_etag = ret.headers["ETag"]
            assert ret_val == cs_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{CAMPAIGN_SCOPES_URL}{DUMMY_ID}",
                json=cs_1_put,
                headers={"If-Match": cs_1_etag},
            )
            assert ret.status_code == 404

            # POST campaign_scope 2
            cs_2 = {
                "name": "Campaign scope 2",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(CAMPAIGN_SCOPES_URL, json=cs_2)
            ret_val = ret.json
            cs_2_id = ret_val.pop("id")
            cs_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            cs_1_put["name"] = cs_2["name"]
            ret = client.put(
                f"{CAMPAIGN_SCOPES_URL}{cs_1_id}",
                json=cs_1_put,
                headers={"If-Match": cs_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                CAMPAIGN_SCOPES_URL, query_string={"name": "Campaign scope 1"}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == cs_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{CAMPAIGN_SCOPES_URL}{DUMMY_ID}", headers={"If-Match": cs_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{CAMPAIGN_SCOPES_URL}{cs_2_id}", headers={"If-Match": cs_2_etag}
            )
            assert ret.status_code == 204

            # DELETE campaign cascade
            ret = client.get(f"{CAMPAIGNS_URL}{campaign_1_id}")
            campaign_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{CAMPAIGNS_URL}{campaign_1_id}", headers={"If-Match": campaign_1_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_campaign_scopes_as_user_api(self, app, users, campaigns, campaign_scopes):

        user_creds = users["Active"]["creds"]
        campaign_1_id = campaigns[0]
        cs_1_id = campaign_scopes[0]
        cs_2_id = campaign_scopes[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(CAMPAIGN_SCOPES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            cs_1 = ret_val[0]
            assert cs_1.pop("id") == cs_1_id

            # GET list with filters
            ret = client.get(
                CAMPAIGN_SCOPES_URL, query_string={"name": "Campaign 1 - Scope 1"}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == cs_1_id
            ret = client.get(
                CAMPAIGN_SCOPES_URL, query_string={"name": "Campaign 2 - Scope 1"}
            )
            assert ret.status_code == 200
            assert not ret.json

            # POST
            cs_3 = {
                "name": "Campaign 1 - Scope 2",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(CAMPAIGN_SCOPES_URL, json=cs_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_1_id}")
            assert ret.status_code == 200
            cs_1_etag = ret.headers["ETag"]

            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_2_id}")
            assert ret.status_code == 403

            # PUT
            cs_1["description"] = "Fantastic campaign"
            cs_1_put = cs_1.copy()
            del cs_1_put["campaign_id"]
            ret = client.put(
                f"{CAMPAIGN_SCOPES_URL}{cs_1_id}",
                json=cs_1_put,
                headers={"If-Match": cs_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{CAMPAIGN_SCOPES_URL}{cs_1_id}", headers={"If-Match": cs_1_etag}
            )
            assert ret.status_code == 403

    def test_campaign_scopes_as_anonym_api(self, app, campaign_scopes):

        cs_1_id = campaign_scopes[0]

        client = app.test_client()

        # GET list
        ret = client.get(CAMPAIGN_SCOPES_URL)
        assert ret.status_code == 401

        # POST
        cs_3 = {
            "name": "Campaign scope 3",
        }
        ret = client.post(CAMPAIGN_SCOPES_URL, json=cs_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_1_id}")
        assert ret.status_code == 401

        # PUT
        cs_1 = {
            "name": "Super campaign scope 1",
        }
        ret = client.put(
            f"{CAMPAIGN_SCOPES_URL}{cs_1_id}",
            json=cs_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{CAMPAIGN_SCOPES_URL}{cs_1_id}", headers={"If-Match": "Dummy-Etag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Timeseries by spaces API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesBySpace

from bemserver_api import AutoSchema, Schema


class TimeseriesBySpaceSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesBySpace

    id = msa.auto_field(dump_only=True)


class TimeseriesBySpaceQueryArgsSchema(Schema):
    space_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()



"""Timeseries properties tests"""
import copy

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_PROPERTIES_URL = "/timeseries_properties/"


class TestTimeseriesDataStatesApi:
    def test_timeseries_properties_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            nb_ts_props = len(ret_val)
            assert nb_ts_props > 0
            nb_float = len([x for x in ret_val if x["value_type"] == "float"])
            nb_bool = len([x for x in ret_val if x["value_type"] == "boolean"])

            # POST
            tsp_1 = {
                "name": "Deutsche Qualität",
                "value_type": "float",
                "unit_symbol": "Qualität unit",
            }
            ret = client.post(TIMESERIES_PROPERTIES_URL, json=tsp_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tsp_1_id = ret_val.pop("id")
            tsp_1_etag = ret.headers["ETag"]
            assert ret_val == tsp_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_PROPERTIES_URL, json=tsp_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_ts_props + 1
            assert tsp_1["name"] in [x["name"] for x in ret_val]

            # GET by id
            ret = client.get(f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tsp_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tsp_1

            # PUT
            tsp_1["name"] = "Qualität"
            tsp_1_put = copy.deepcopy(tsp_1)
            del tsp_1_put["value_type"]
            ret = client.put(
                f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}",
                json=tsp_1_put,
                headers={"If-Match": tsp_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            tsp_1_etag = ret.headers["ETag"]
            assert ret_val == tsp_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_PROPERTIES_URL}{DUMMY_ID}",
                json=tsp_1_put,
                headers={"If-Match": tsp_1_etag},
            )
            assert ret.status_code == 404

            # POST TSP 2
            tsp_2 = {
                "name": "Autentica qualità",
                "value_type": "float",
            }
            ret = client.post(TIMESERIES_PROPERTIES_URL, json=tsp_2)
            ret_val = ret.json
            tsp_2_id = ret_val.pop("id")
            tsp_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            tsp_2_put = copy.deepcopy(tsp_2)
            del tsp_2_put["value_type"]
            tsp_2_put["name"] = "Qualität"
            ret = client.put(
                f"{TIMESERIES_PROPERTIES_URL}{tsp_2_id}",
                json=tsp_2_put,
                headers={"If-Match": tsp_2_etag},
            )
            assert ret.status_code == 409

            # GET list with filters
            ret = client.get(
                TIMESERIES_PROPERTIES_URL, query_string={"name": "Qualität"}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == tsp_1_id
            ret = client.get(
                TIMESERIES_PROPERTIES_URL,
                query_string={"value_type": "float"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_float + 2
            ret = client.get(
                TIMESERIES_PROPERTIES_URL,
                query_string={"value_type": "boolean"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_bool
            ret = client.get(
                TIMESERIES_PROPERTIES_URL,
                query_string={"unit_symbol": "Qualität unit"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ret = client.get(
                TIMESERIES_PROPERTIES_URL,
                query_string={"unit_symbol": "Qualità unit"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}",
                headers={"If-Match": tsp_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{TIMESERIES_PROPERTIES_URL}{tsp_2_id}",
                headers={"If-Match": tsp_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(TIMESERIES_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_ts_props

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}")
            assert ret.status_code == 404

    def test_timeseries_properties_as_user_api(self, app, users):

        tds_1_id = 1

        creds = users["Active"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            nb_ts_props = len(ret_val)
            assert nb_ts_props > 0

            # POST
            tds = {
                "name": "Frequency (test)",
                "value_type": "integer",
            }
            ret = client.post(TIMESERIES_PROPERTIES_URL, json=tds)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_PROPERTIES_URL}{tds_1_id}")
            assert ret.status_code == 200
            tds_1_etag = ret.headers["ETag"]
            ret_val = ret.json
            ret_val.pop("id")
            tds_1 = ret_val

            # PUT
            tds_1_put = copy.deepcopy(tds_1)
            del tds_1_put["value_type"]
            tds_1_put["name"] = "Qualität"
            ret = client.put(
                f"{TIMESERIES_PROPERTIES_URL}{tds_1_id}",
                json=tds_1_put,
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_PROPERTIES_URL}{tds_1_id}",
                headers={"If-Match": tds_1_etag},
            )
            assert ret.status_code == 403

    def test_timeseries_properties_as_anonym_api(
        self, app, users, timeseries_properties
    ):

        tsp_1_id = timeseries_properties[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_PROPERTIES_URL)
        assert ret.status_code == 401

        # POST
        tds = {
            "name": "Frequency (test)",
        }
        ret = client.post(TIMESERIES_PROPERTIES_URL, json=tds)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}")
        assert ret.status_code == 401

        # PUT
        ret = client.put(
            f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}",
            json=tds,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_PROPERTIES_URL}{tsp_1_id}",
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Spaces API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Space

from bemserver_api import AutoSchema, Schema, SortField


class SpaceSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Space

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    storey_id = ma.fields.Int(required=True)


class SpacePutSchema(SpaceSchema):
    class Meta(SpaceSchema.Meta):
        exclude = ("storey_id",)


class SpaceQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    campaign_id = ma.fields.Int()
    site_id = ma.fields.Int()
    building_id = ma.fields.Int()
    storey_id = ma.fields.Int()
    ifc_id = ma.fields.String()



"""Structural element properties routes tests"""
import copy

from tests.common import AuthHeader


DUMMY_ID = "69"

STRUCTURAL_ELEMENT_PROPERTIES_URL = "/structural_element_properties/"


class TestStructuralElementPropertiesApi:
    def test_structural_element_properties_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            sep_1 = {
                "name": "Area",
                "value_type": "integer",
            }
            ret = client.post(STRUCTURAL_ELEMENT_PROPERTIES_URL, json=sep_1)
            assert ret.status_code == 201
            ret_val = ret.json
            sep_1_id = ret_val.pop("id")
            sep_1_etag = ret.headers["ETag"]
            assert ret_val == sep_1

            # POST violating unique constraint
            ret = client.post(STRUCTURAL_ELEMENT_PROPERTIES_URL, json=sep_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == sep_1_id

            # GET by id
            ret = client.get(f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == sep_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == sep_1

            # PUT
            sep_1["description"] = "Fantastic sep"
            sep_1_put = copy.deepcopy(sep_1)
            del sep_1_put["value_type"]
            ret = client.put(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
                json=sep_1_put,
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            sep_1_etag = ret.headers["ETag"]
            assert ret_val == sep_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{DUMMY_ID}",
                json=sep_1_put,
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            sep_2 = {
                "name": "Volume",
                "unit_symbol": "L",
            }
            ret = client.post(STRUCTURAL_ELEMENT_PROPERTIES_URL, json=sep_2)
            assert ret.status_code == 201
            ret_val = ret.json
            assert ret_val["value_type"] == "string"
            sep_2_id = ret_val.pop("id")
            sep_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            sep_2_put = copy.deepcopy(sep_2)
            sep_2_put["name"] = sep_1["name"]
            ret = client.put(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_2_id}",
                json=sep_2_put,
                headers={"If-Match": sep_2_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                STRUCTURAL_ELEMENT_PROPERTIES_URL, query_string={"name": "Area"}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == sep_1_id
            ret = client.get(
                STRUCTURAL_ELEMENT_PROPERTIES_URL,
                query_string={"value_type": "integer"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == sep_1_id
            ret = client.get(
                STRUCTURAL_ELEMENT_PROPERTIES_URL,
                query_string={"value_type": "boolean"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(
                STRUCTURAL_ELEMENT_PROPERTIES_URL,
                query_string={"unit_symbol": "L"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ret = client.get(
                STRUCTURAL_ELEMENT_PROPERTIES_URL,
                query_string={"unit_symbol": "kWh"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{DUMMY_ID}",
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_2_id}",
                headers={"If-Match": sep_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}")
            assert ret.status_code == 404

    def test_structural_element_as_user_api(
        self, app, users, structural_element_properties
    ):

        user_creds = users["Active"]["creds"]
        sep_1_id = structural_element_properties[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 4

            # POST
            sep_3 = {
                "name": "Temperature setpoint",
                "value_type": "float",
            }
            ret = client.post(STRUCTURAL_ELEMENT_PROPERTIES_URL, json=sep_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}")
            assert ret.status_code == 200
            sep_1 = ret.json
            del sep_1["id"]
            sep_1_etag = ret.headers["ETag"]

            # PUT
            sep_1_put = copy.deepcopy(sep_1)
            del sep_1_put["value_type"]
            sep_1_put["description"] = "Fantastic sep"
            ret = client.put(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
                json=sep_1_put,
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
                headers={"If-Match": sep_1_etag},
            )
            assert ret.status_code == 403

    def test_structural_element_as_anonym_api(self, app, structural_element_properties):

        sep_1_id = structural_element_properties[0]

        client = app.test_client()

        # GET list
        ret = client.get(STRUCTURAL_ELEMENT_PROPERTIES_URL)
        assert ret.status_code == 401

        # POST
        sep_3 = {
            "name": "Temperature setpoint",
            "value_type": "float",
        }
        ret = client.post(STRUCTURAL_ELEMENT_PROPERTIES_URL, json=sep_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}")
        assert ret.status_code == 401

        # PUT
        sep_1 = {
            "name": "Gross surface",
        }
        ret = client.put(
            f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
            json=sep_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{STRUCTURAL_ELEMENT_PROPERTIES_URL}{sep_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Timeseries data states API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesDataState

from bemserver_api import AutoSchema


class TimeseriesDataStateSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesDataState

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))



"""Event levels resources"""

from flask.views import MethodView

from bemserver_core.model import EventLevel

from bemserver_api import Blueprint

from .schemas import EventLevelSchema


blp = Blueprint(
    "EventLevel",
    __name__,
    url_prefix="/event_levels",
    description="Operations on event levels",
)


@blp.route("/")
class EventLevelsViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EventLevelSchema(many=True))
    def get(self):
        """List event levels"""
        return EventLevel.get()



"""Space resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Space

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import SpaceSchema, SpacePutSchema, SpaceQueryArgsSchema


blp = Blueprint(
    "Space", __name__, url_prefix="/spaces", description="Operations on spaces"
)


@blp.route("/")
class SpaceViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(SpaceQueryArgsSchema, location="query")
    @blp.response(200, SpaceSchema(many=True))
    def get(self, args):
        """List spaces"""
        return Space.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(SpaceSchema)
    @blp.response(201, SpaceSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new space"""
        item = Space.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class SpaceByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, SpaceSchema)
    def get(self, item_id):
        """Get space by ID"""
        item = Space.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(SpacePutSchema)
    @blp.response(200, SpaceSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing space"""
        item = Space.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, SpaceSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a space"""
        item = Space.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, SpaceSchema)
        item.delete()
        db.session.commit()



"""Event categories tests"""
from tests.common import AuthHeader


DUMMY_ID = "69"

EVENT_CATEGORIES_URL = "/event_categories/"


class TestEventCategoriesApi:
    def test_event_categories_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(EVENT_CATEGORIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            nb_categ = len(ret_val)

            # POST
            event_categ_1 = {
                "name": "Custom category 1",
                "description": "Short description",
            }
            ret = client.post(EVENT_CATEGORIES_URL, json=event_categ_1)
            assert ret.status_code == 201
            ret_val = ret.json
            event_categ_1_id = ret_val.pop("id")
            event_categ_1_etag = ret.headers["ETag"]

            # POST violating unique constraint
            ret = client.post(EVENT_CATEGORIES_URL, json=event_categ_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(EVENT_CATEGORIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_categ + 1

            # GET by id
            ret = client.get(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == event_categ_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == event_categ_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{EVENT_CATEGORIES_URL}{DUMMY_ID}",
                json=event_categ_1,
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 404

            # PUT
            event_categ_1["description"] = "Longer description"
            ret = client.put(
                f"{EVENT_CATEGORIES_URL}{event_categ_1_id}",
                json=event_categ_1,
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            event_categ_1_etag = ret.headers["ETag"]

            # POST
            event_categ_2 = {
                "name": "Custom category 2",
            }
            ret = client.post(EVENT_CATEGORIES_URL, json=event_categ_2)
            ret_val = ret.json

            # GET list
            ret = client.get(EVENT_CATEGORIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_categ + 2

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{EVENT_CATEGORIES_URL}{DUMMY_ID}",
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{EVENT_CATEGORIES_URL}{event_categ_1_id}",
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(EVENT_CATEGORIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_categ + 1

            # GET by id -> 404
            ret = client.get(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}")
            assert ret.status_code == 404

    def test_event_categories_as_user_api(self, app, users, event_categories):

        user_creds = users["Active"]["creds"]
        event_categ_1_id = event_categories[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(EVENT_CATEGORIES_URL)
            assert ret.status_code == 200

            # POST
            event_categ = {
                "name": "Custom category 1",
                "description": "Short description",
            }
            ret = client.post(EVENT_CATEGORIES_URL, json=event_categ)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}")
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            event_categ_1 = ret_val
            event_categ_1_etag = ret.headers["ETag"]

            # PUT
            ret = client.put(
                f"{EVENT_CATEGORIES_URL}{event_categ_1_id}",
                json=event_categ_1,
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{EVENT_CATEGORIES_URL}{event_categ_1_id}",
                headers={"If-Match": event_categ_1_etag},
            )
            assert ret.status_code == 403

    def test_event_categories_as_anonym_api(self, app, event_categories):

        event_categ_1_id = event_categories[0]

        client = app.test_client()

        # GET list
        ret = client.get(EVENT_CATEGORIES_URL)
        assert ret.status_code == 401

        # POST
        event_categ = {
            "name": "Custom category 1",
            "description": "Short description",
        }
        ret = client.post(EVENT_CATEGORIES_URL, json=event_categ)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}")
        assert ret.status_code == 401

        # PUT
        ret = client.put(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}", json=event_categ)

        # DELETE
        ret = client.delete(f"{EVENT_CATEGORIES_URL}{event_categ_1_id}")
        assert ret.status_code == 401



import marshmallow as ma

import pytest

from bemserver_api.extensions.ma_fields import Timezone


class TestMaFields:
    def test_ma_fields_timezone(self):
        field = Timezone()
        assert field.deserialize("UTC") == "UTC"
        assert field.deserialize("Europe/Paris") == "Europe/Paris"
        with pytest.raises(ma.ValidationError):
            field.deserialize("Dordogne/Boulazac")



"""Campaigns routes tests"""
import datetime as dt

import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

CAMPAIGNS_URL = "/campaigns/"


class TestCampaignsApi:
    def test_campaigns_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(CAMPAIGNS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            campaign_1 = {
                "name": "Campaign 1",
                "start_time": (
                    dt.datetime(2012, 9, 4, tzinfo=dt.timezone.utc).isoformat()
                ),
                "end_time": (
                    dt.datetime(2017, 9, 29, tzinfo=dt.timezone.utc).isoformat()
                ),
                "timezone": "Europe/Paris",
            }
            ret = client.post(CAMPAIGNS_URL, json=campaign_1)
            assert ret.status_code == 201
            ret_val = ret.json
            campaign_1_id = ret_val.pop("id")
            campaign_1_etag = ret.headers["ETag"]
            assert ret_val == campaign_1

            # POST violating unique constraint
            ret = client.post(CAMPAIGNS_URL, json=campaign_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == campaign_1_id

            # GET by id
            ret = client.get(f"{CAMPAIGNS_URL}{campaign_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == campaign_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == campaign_1

            # PUT
            campaign_1["description"] = "Fantastic campaign"
            ret = client.put(
                f"{CAMPAIGNS_URL}{campaign_1_id}",
                json=campaign_1,
                headers={"If-Match": campaign_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            campaign_1_etag = ret.headers["ETag"]
            assert ret_val == campaign_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{CAMPAIGNS_URL}{DUMMY_ID}",
                json=campaign_1,
                headers={"If-Match": campaign_1_etag},
            )
            assert ret.status_code == 404

            # POST campaign 2
            campaign_2 = {
                "name": "Campaign 2",
                "start_time": (
                    dt.datetime(2016, 4, 8, tzinfo=dt.timezone.utc).isoformat()
                ),
                "end_time": (
                    dt.datetime(2019, 9, 20, tzinfo=dt.timezone.utc).isoformat()
                ),
                "timezone": "UTC",
            }
            ret = client.post(CAMPAIGNS_URL, json=campaign_2)
            ret_val = ret.json
            campaign_2_id = ret_val.pop("id")
            campaign_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            campaign_2["name"] = campaign_1["name"]
            ret = client.put(
                f"{CAMPAIGNS_URL}{campaign_2_id}",
                json=campaign_2,
                headers={"If-Match": campaign_2_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(CAMPAIGNS_URL, query_string={"name": "Campaign 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == campaign_1_id

            # GET sorted list
            ret = client.get(CAMPAIGNS_URL, query_string={"sort": "name"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == campaign_1_id
            assert ret_val[1]["id"] == campaign_2_id
            ret = client.get(CAMPAIGNS_URL, query_string={"sort": "+name"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == campaign_1_id
            assert ret_val[1]["id"] == campaign_2_id
            ret = client.get(CAMPAIGNS_URL, query_string={"sort": "-name"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == campaign_2_id
            assert ret_val[1]["id"] == campaign_1_id

            # GET list using "in_name"
            ret = client.get(CAMPAIGNS_URL, query_string={"in_name": "not"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(CAMPAIGNS_URL, query_string={"in_name": "paign"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == campaign_1_id
            ret = client.get(CAMPAIGNS_URL, query_string={"in_name": "paign 2"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == campaign_2_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{CAMPAIGNS_URL}{DUMMY_ID}", headers={"If-Match": campaign_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{CAMPAIGNS_URL}{campaign_1_id}", headers={"If-Match": campaign_1_etag}
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{CAMPAIGNS_URL}{campaign_2_id}", headers={"If-Match": campaign_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{CAMPAIGNS_URL}{campaign_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_campaigns_as_user_api(self, app, users, campaigns):

        user_creds = users["Active"]["creds"]
        campaign_1_id, campaign_2_id = campaigns

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            campaign_1 = ret_val[0]
            assert campaign_1.pop("id") == campaign_1_id

            # GET list with filters
            ret = client.get(CAMPAIGNS_URL, query_string={"name": "Campaign 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == campaign_1_id
            ret = client.get(CAMPAIGNS_URL, query_string={"name": "Campaign 2"})
            assert ret.status_code == 200
            assert not ret.json

            # GET list using "in_name"
            ret = client.get(CAMPAIGNS_URL, query_string={"in_name": "not"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(CAMPAIGNS_URL, query_string={"in_name": "paign"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == campaign_1_id

            # POST
            campaign_3 = {
                "name": "Campaign 3",
                "start_time": (
                    dt.datetime(2012, 9, 4, tzinfo=dt.timezone.utc).isoformat()
                ),
                "end_time": (
                    dt.datetime(2017, 9, 29, tzinfo=dt.timezone.utc).isoformat()
                ),
                "timezone": "UTC",
            }
            ret = client.post(CAMPAIGNS_URL, json=campaign_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{CAMPAIGNS_URL}{campaign_1_id}")
            assert ret.status_code == 200
            campaign_1_etag = ret.headers["ETag"]

            ret = client.get(f"{CAMPAIGNS_URL}{campaign_2_id}")
            assert ret.status_code == 403

            # PUT
            campaign_1["description"] = "Fantastic campaign"
            ret = client.put(
                f"{CAMPAIGNS_URL}{campaign_1_id}",
                json=campaign_1,
                headers={"If-Match": campaign_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{CAMPAIGNS_URL}{campaign_1_id}", headers={"If-Match": campaign_1_etag}
            )
            assert ret.status_code == 403

    def test_campaigns_as_anonym_api(self, app, campaigns):

        campaign_1_id, _ = campaigns

        client = app.test_client()

        # GET list
        ret = client.get(CAMPAIGNS_URL)
        assert ret.status_code == 401

        # POST
        campaign_3 = {
            "name": "Campaign 3",
            "start_time": (dt.datetime(2012, 9, 4, tzinfo=dt.timezone.utc).isoformat()),
            "end_time": (dt.datetime(2017, 9, 29, tzinfo=dt.timezone.utc).isoformat()),
            "timezone": "UTC",
        }
        ret = client.post(CAMPAIGNS_URL, json=campaign_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{CAMPAIGNS_URL}{campaign_1_id}")
        assert ret.status_code == 401

        # PUT
        campaign_1 = {
            "name": "Super Campaign 1",
        }
        ret = client.put(
            f"{CAMPAIGNS_URL}{campaign_1_id}",
            json=campaign_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{CAMPAIGNS_URL}{campaign_1_id}", headers={"If-Match": "Dummy-Etag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Energy consumption timeseries by buildings resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import EnergyConsumptionTimeseriesByBuilding

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    EnergyConsumptionTimeseriesByBuildingSchema,
    EnergyConsumptionTimeseriesByBuildingQueryArgsSchema,
)


blp = Blueprint(
    "EnergyConsumptionTimeseriesByBuilding",
    __name__,
    url_prefix="/energy_consumption_timeseries_by_buildings",
    description="Operations on energy consumption timeseries x building associations",
)


@blp.route("/")
class EnergyConsumptionTimeseriesByBuildingViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(
        EnergyConsumptionTimeseriesByBuildingQueryArgsSchema, location="query"
    )
    @blp.response(200, EnergyConsumptionTimeseriesByBuildingSchema(many=True))
    def get(self, args):
        """List energy consumption timeseries x building associations"""
        return EnergyConsumptionTimeseriesByBuilding.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(EnergyConsumptionTimeseriesByBuildingSchema)
    @blp.response(201, EnergyConsumptionTimeseriesByBuildingSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new energy consumption timeseries x building association"""
        item = EnergyConsumptionTimeseriesByBuilding.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class EnergyConsumptionTimeseriesByBuildingByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EnergyConsumptionTimeseriesByBuildingSchema)
    def get(self, item_id):
        """Get energy consumption timeseries x building association ID"""
        item = EnergyConsumptionTimeseriesByBuilding.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(EnergyConsumptionTimeseriesByBuildingSchema)
    @blp.response(200, EnergyConsumptionTimeseriesByBuildingSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing energy consumption timeseries x building association"""
        item = EnergyConsumptionTimeseriesByBuilding.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EnergyConsumptionTimeseriesByBuildingSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a energy consumption timeseries x building association"""
        item = EnergyConsumptionTimeseriesByBuilding.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EnergyConsumptionTimeseriesByBuildingSchema)
        item.delete()
        db.session.commit()



"""STCheckMissingByCampaigns routes tests"""
from copy import deepcopy

import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL = "/st_check_missings_by_campaigns/"


class TestST_CheckMissingByCampaignsApi:
    def test_st_check_missings_by_campaigns_api(self, app, users, campaigns):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            st_cbc_1 = {"campaign_id": campaign_1_id}
            ret = client.post(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL, json=st_cbc_1)
            assert ret.status_code == 201
            ret_val = ret.json
            st_cbc_1_id = ret_val.pop("id")
            st_cbc_1_etag = ret.headers["ETag"]
            st_cbc_1_expected = deepcopy(st_cbc_1)
            st_cbc_1_expected["is_enabled"] = True
            assert ret_val == st_cbc_1_expected

            # POST violating unique constraint
            ret = client.post(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL, json=st_cbc_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbc_1_id

            # GET by id
            ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == st_cbc_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == st_cbc_1_expected

            # PUT
            st_cbc_1_expected["is_enabled"] = False
            ret = client.put(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
                json={"is_enabled": False},
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            st_cbc_1_etag = ret.headers["ETag"]
            assert ret_val == st_cbc_1_expected

            # PUT wrong ID -> 404
            ret = client.put(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{DUMMY_ID}",
                json={"is_enabled": False},
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 404

            # POST campaign 2
            st_cbc_2 = {"campaign_id": campaign_2_id}
            ret = client.post(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL, json=st_cbc_2)
            ret_val = ret.json
            st_cbc_2_id = ret_val.pop("id")
            st_cbc_2_etag = ret.headers["ETag"]

            # GET list
            ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL,
                query_string={"campaign_id": campaign_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbc_1_id

            # GET list (full), sort by campaign name
            ret = client.get(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}full",
                query_string={"sort": "+campaign_name"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == st_cbc_1_id
            assert ret_val[0]["campaign_id"] == campaign_1_id
            assert not ret_val[0]["is_enabled"]
            assert ret_val[1]["id"] == st_cbc_2_id
            assert ret_val[1]["campaign_id"] == campaign_2_id
            assert ret_val[1]["is_enabled"]

            # GET list (full), sort by campaign name and filter by state
            ret = client.get(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}full",
                query_string={"is_enabled": True, "sort": "+campaign_name"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbc_2_id
            assert ret_val[0]["campaign_id"] == campaign_2_id
            assert ret_val[0]["is_enabled"]

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{DUMMY_ID}",
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_2_id}",
                headers={"If-Match": st_cbc_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_st_check_missings_by_campaigns_as_user_api(
        self, app, users, campaigns, st_check_missings_by_campaigns
    ):

        user_creds = users["Active"]["creds"]
        campaign_1_id = campaigns[0]
        st_cbc_1_id = st_check_missings_by_campaigns[0]
        st_cbc_2_id = st_check_missings_by_campaigns[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            st_cbc_1 = ret_val[0]
            assert st_cbc_1["id"] == st_cbc_1_id

            # POST
            st_cbc_3 = {"campaign_id": campaign_1_id}
            ret = client.post(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL, json=st_cbc_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}")
            assert ret.status_code == 200
            ret_val = ret.json
            assert ret_val["id"] == st_cbc_1_id
            st_cbc_1_etag = ret.headers["ETag"]
            ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_2_id}")
            assert ret.status_code == 403

            # PUT
            ret = client.put(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
                json={"is_enabled": False},
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
                headers={"If-Match": st_cbc_1_etag},
            )
            assert ret.status_code == 403

            # GET list (full)
            ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}full")
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbc_1_id

            # GET list (full), filter by state is_enabled
            ret = client.get(
                f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}full",
                query_string={"is_enabled": False},
            )
            assert ret.status_code == 200
            assert len(ret.json) == 0

    def test_st_check_missings_by_campaigns_as_anonym_api(
        self, app, campaigns, st_check_missings_by_campaigns
    ):

        campaign_1_id = campaigns[0]
        st_cbc_1_id = st_check_missings_by_campaigns[0]

        client = app.test_client()

        # GET list
        ret = client.get(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL)
        assert ret.status_code == 401

        # POST
        st_cbc_3 = {"campaign_id": campaign_1_id}
        ret = client.post(ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL, json=st_cbc_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}")
        assert ret.status_code == 401

        # PUT
        ret = client.put(
            f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
            json={"is_enabled": False},
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}{st_cbc_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # GET list (full)
        ret = client.get(f"{ST_CHECK_MISSINGS_BY_CAMPAIGNS_URL}full")
        assert ret.status_code == 401



"""Buildings API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Building

from bemserver_api import AutoSchema, Schema, SortField


class BuildingSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Building

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    site_id = ma.fields.Int(required=True)


class BuildingPutSchema(BuildingSchema):
    class Meta(BuildingSchema.Meta):
        exclude = ("site_id",)


class BuildingQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    campaign_id = ma.fields.Int()
    site_id = ma.fields.Int()
    ifc_id = ma.fields.String()



"""About API schemas"""
import marshmallow as ma

from bemserver_api import Schema


class VersionsSchema(Schema):
    bemserver_core = ma.fields.Str()
    bemserver_api = ma.fields.Str()


class AboutSchema(Schema):
    versions = ma.fields.Nested(VersionsSchema)



"""Space properties routes tests"""
from tests.common import AuthHeader


DUMMY_ID = "69"

SPACE_PROPERTIES_URL = "/space_properties/"


class TestSpacePropertiesApi:
    def test_space_properties_api(self, app, users, structural_element_properties):

        creds = users["Chuck"]["creds"]
        sep_1_id = structural_element_properties[0]
        sep_2_id = structural_element_properties[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(SPACE_PROPERTIES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            space_p_1 = {
                "structural_element_property_id": sep_1_id,
            }
            ret = client.post(SPACE_PROPERTIES_URL, json=space_p_1)
            assert ret.status_code == 201
            ret_val = ret.json
            space_p_1_id = ret_val.pop("id")
            space_p_1_etag = ret.headers["ETag"]
            sep = ret_val.pop("structural_element_property")
            assert sep == {"name": "Area", "value_type": "integer"}
            assert ret_val == space_p_1

            # POST violating unique constraint
            ret = client.post(SPACE_PROPERTIES_URL, json=space_p_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(SPACE_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == space_p_1_id

            # GET by id
            ret = client.get(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == space_p_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            ret_val.pop("structural_element_property")
            assert ret_val == space_p_1

            # POST sep 2
            space_p_2 = {
                "structural_element_property_id": sep_2_id,
            }
            ret = client.post(SPACE_PROPERTIES_URL, json=space_p_2)
            ret_val = ret.json
            space_p_2_id = ret_val.pop("id")

            # GET list
            ret = client.get(SPACE_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                SPACE_PROPERTIES_URL,
                query_string={"structural_element_property_id": sep_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == space_p_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(f"{SPACE_PROPERTIES_URL}{DUMMY_ID}")
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
            assert ret.status_code == 204
            ret = client.delete(f"{SPACE_PROPERTIES_URL}{space_p_2_id}")
            assert ret.status_code == 204

            # GET list
            ret = client.get(SPACE_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
            assert ret.status_code == 404

    def test_space_properties_as_user_api(
        self, app, users, structural_element_properties, space_properties
    ):

        user_creds = users["Active"]["creds"]
        sep_1_id = structural_element_properties[0]
        space_p_1_id = space_properties[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(SPACE_PROPERTIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            space_p_1 = ret_val[0]
            assert space_p_1.pop("id") == space_p_1_id

            # POST
            space_p_3 = {
                "structural_element_property_id": sep_1_id,
            }
            ret = client.post(SPACE_PROPERTIES_URL, json=space_p_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
            assert ret.status_code == 200

            # DELETE
            ret = client.delete(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
            assert ret.status_code == 403

    def test_space_properties_as_anonym_api(
        self, app, structural_element_properties, space_properties
    ):

        sep_1_id = structural_element_properties[0]
        space_p_1_id = space_properties[0]

        client = app.test_client()

        # GET list
        ret = client.get(SPACE_PROPERTIES_URL)
        assert ret.status_code == 401

        # POST
        space_p_3 = {
            "structural_element_property_id": sep_1_id,
        }
        ret = client.post(SPACE_PROPERTIES_URL, json=space_p_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(f"{SPACE_PROPERTIES_URL}{space_p_1_id}")
        assert ret.status_code == 401



"""Timeseries API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Timeseries

from bemserver_api import AutoSchema, Schema, SortField


class TimeseriesSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Timeseries

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    campaign_id = ma.fields.Int(required=True)
    campaign_scope_id = ma.fields.Int(required=True)
    unit_symbol = msa.auto_field()


class TimeseriesPutSchema(TimeseriesSchema):
    class Meta(TimeseriesSchema.Meta):
        exclude = ("campaign_id", "campaign_scope_id")


class TimeseriesQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    in_name = ma.fields.Str()
    unit_symbol = ma.fields.Str()
    campaign_id = ma.fields.Int()
    campaign_scope_id = ma.fields.Int()
    user_id = ma.fields.Int()
    site_id = ma.fields.Int()
    building_id = ma.fields.Int()
    storey_id = ma.fields.Int()
    space_id = ma.fields.Int()
    zone_id = ma.fields.Int()



"""Zone property data API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import ZonePropertyData

from bemserver_api import AutoSchema, Schema


class ZonePropertyDataSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = ZonePropertyData

    id = msa.auto_field(dump_only=True)


class ZonePropertyDataQueryArgsSchema(Schema):
    zone_id = ma.fields.Int()
    zone_property_id = ma.fields.Int()



"""Site property data resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import SitePropertyData
from bemserver_core.exceptions import PropertyTypeInvalidError

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    SitePropertyDataSchema,
    SitePropertyDataQueryArgsSchema,
)


blp = Blueprint(
    "SitePropertyData",
    __name__,
    url_prefix="/site_property_data",
    description="Operations on site property data",
)


@blp.route("/")
class SitePropertyDataViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(SitePropertyDataQueryArgsSchema, location="query")
    @blp.response(200, SitePropertyDataSchema(many=True))
    def get(self, args):
        """List site property data"""
        return SitePropertyData.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(SitePropertyDataSchema)
    @blp.response(201, SitePropertyDataSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new site property data"""
        item = SitePropertyData.new(**new_item)
        try:
            db.session.commit()
        except PropertyTypeInvalidError:
            abort(422, errors={"json": {"value": ["Invalid type."]}})
        return item


@blp.route("/<int:item_id>")
class SitePropertyDataByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, SitePropertyDataSchema)
    def get(self, item_id):
        """Get site property data by ID"""
        item = SitePropertyData.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(SitePropertyDataSchema)
    @blp.response(200, SitePropertyDataSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing site property data"""
        item = SitePropertyData.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, SitePropertyDataSchema)
        item.update(**new_item)
        try:
            db.session.commit()
        except PropertyTypeInvalidError:
            abort(422, errors={"json": {"value": ["Invalid type."]}})
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a site property data"""
        item = SitePropertyData.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, SitePropertyDataSchema)
        item.delete()
        db.session.commit()



"""Timeseries by events API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesByEvent

from bemserver_api import AutoSchema, Schema


class TimeseriesByEventSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesByEvent

    id = msa.auto_field(dump_only=True)


class TimeseriesByEventQueryArgsSchema(Schema):
    event_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()



"""Building resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Building

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import BuildingSchema, BuildingPutSchema, BuildingQueryArgsSchema


blp = Blueprint(
    "Building", __name__, url_prefix="/buildings", description="Operations on buildings"
)


@blp.route("/")
class BuildingViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(BuildingQueryArgsSchema, location="query")
    @blp.response(200, BuildingSchema(many=True))
    def get(self, args):
        """List buildings"""
        return Building.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(BuildingSchema)
    @blp.response(201, BuildingSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new building"""
        item = Building.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class BuildingByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, BuildingSchema)
    def get(self, item_id):
        """Get building by ID"""
        item = Building.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(BuildingPutSchema)
    @blp.response(200, BuildingSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing building"""
        item = Building.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, BuildingSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a building"""
        item = Building.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, BuildingSchema)
        item.delete()
        db.session.commit()
